"""
uqid.py version 0.3.0
https://github.com/denis-ryzhkov/uqidpy

Copyright (C) 2015-2018 by Denis Ryzhkov <denisr@denisr.com>
MIT License, see http://opensource.org/licenses/MIT
"""

### import

from datetime import datetime
from random import choice

try:
    xrange
except NameError:
    xrange = range

### chars

digits = '0123456789'
base62 = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'

### uqid

def uqid(length=64, chars=base62):
    return ''.join(choice(chars) for i in xrange(length))

### dtid

_dt_format = '%Y%m%d%H%M%S%f'
_dt_length = 20

def dtid(length=24, chars=base62):
    s = datetime.utcnow().strftime(_dt_format)
    if length > _dt_length:
        return s + uqid(length - _dt_length, chars)
    return s[:length]

### datetime_from_dtid

def datetime_from_dtid(s):
    return datetime.strptime(s[:_dt_length], _dt_format)

### tests

def tests():

    ### import

    from time import time
    from uuid import uuid4

    try:
        from bson import ObjectId
    except ImportError:
        ObjectId = None

    ### samples

    print('\nSAMPLES:')
    length = 24

    print('str(uuid4()): {}'.format(uuid4()))
    print('len(^): {}'.format(len(str(uuid4()))))
    print('len(v): {}'.format(length))

    if ObjectId:
        print('str(ObjectId()): {}'.format(str(ObjectId())))

    print('uqid({}): {}'.format(length, uqid(length)))
    print('dtid({}): {}'.format(length, dtid(length)))

    assert len(uqid(length)) == length
    assert len(dtid(length)) == length
    assert (datetime.utcnow() - datetime_from_dtid(dtid())).total_seconds() < 0.1

    N = 1000*1000
    print('Iterations: {}'.format(N))

    ### seconds

    print('\nSECONDS:')

    start = time()
    for _ in xrange(N):
        str(uuid4())
    print('str(uuid4()) seconds: {:.6f}'.format(time() - start))

    if ObjectId:
        start = time()
        for _ in xrange(N):
            str(ObjectId())
        print('str(ObjectId()) seconds: {:.6f}'.format(time() - start))

    start = time()
    for _ in xrange(N):
        uqid(length)
    print('uqid({}) seconds: {:.6f}'.format(length, time() - start))

    start = time()
    for _ in xrange(N):
        dtid(length)
    print('dtid({}) seconds: {:.6f}'.format(length, time() - start))

    ### duplicates

    print('\nDUPLICATES:')

    U = len(set(str(uuid4()) for _ in xrange(N)))
    print('str(uuid4()) duplicates: {}'.format(N - U))

    if ObjectId:
        U = len(set(str(ObjectId()) for _ in xrange(N)))
        print('str(ObjectId()) duplicates: {}'.format(N - U))

    U = len(set(uqid(length) for _ in xrange(N)))
    print('uqid({}) duplicates: {}'.format(length, N - U))

    U = len(set(dtid(length) for _ in xrange(N)))
    print('dtid({}) duplicates: {}'.format(length, N - U))

if __name__ == '__main__':
    tests()



from setuptools import setup

setup(
    name='uqid',
    version='0.3.0',
    description='Unique id from random chars and optional date-time.',
    long_description='''
uqid::

    pip install uqid
    from uqid import uqid, digits

    print(uqid()) # e7ljCIs4grYZ9tOv6bpX3TQIA7FyA23umzGjDcIjHff2XwF2ziAw56B7S3EdOuCA
    print(uqid(4, digits)) # 4982

* Returns unique id from random chars.
* Default is 64 chars 0-9a-zA-Z aka base62.

dtid::

    pip install uqid
    from uqid import dtid
    print(dtid())

    YYYYMMDDhhmmssmicrosrnd4
    20161231235959123456uN1q

* 24 bytes by default - as str(ObjectId()).
* But human-friendly - unlike ObjectId.
* Microsecond-precise UTC timestamp - unlike second-precise ObjectId().
* May extract datetime_from_dtid().
* May be used as a log timestamp as is.
* Alphabetically ordered.
* Safe chars 0-9a-zA-Z - unlike binary ObjectId() leading to "ValueError: Invalid UTF-8 sequence length" if added to JSON without str().
* base62^4 is over 14M uniques per microsecond.
* dtid(24) is 2x faster than uqid(24) but slower than str(ObjectId())

tests::

    python2 uqid.py
    python3 uqid.py

    SAMPLES:
    str(uuid4()): 3bb25137-2422-4d06-bd12-5a9de2ea2bf6
    len(^): 36
    len(v): 24
    str(ObjectId()): 5acdcd1f9c591516ae800615
    uqid(24): UFRFQVCtYu6rtQq8LD33XjI3
    dtid(24): 20180411085351268116tHse
    Iterations: 1000000

    python2 SECONDS:
    str(uuid4()) seconds: 11.825964
    str(ObjectId()) seconds: 2.424702
    uqid(24) seconds: 9.701306
    dtid(24) seconds: 5.357983

    python3 SECONDS:
    str(uuid4()) seconds: 5.826098
    str(ObjectId()) seconds: 3.192565
    uqid(24) seconds: 17.769653
    dtid(24) seconds: 9.588677

    DUPLICATES:
    str(uuid4()) duplicates: 0
    str(ObjectId()) duplicates: 0
    uqid(24) duplicates: 0
    dtid(24) duplicates: 0

''',
    url='https://github.com/denis-ryzhkov/uqidpy',
    author='Denis Ryzhkov',
    author_email='denisr@denisr.com',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    py_modules=['uqid'],
)
